#!/usr/bin/env python2.4

# text-based supermemox.  This version schedules the cards
# based on time directly, rather than going through probability 
# curves first.  Currently the datastore path is hardcoded in the UI
# near the bottom of the file.
#
#
#   Robert W. Brewer <rwb123@gmail.com>
#   Copyright (C) 2004 Robert W. Brewer
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  

import os
import pickle
import re
import heapq
import time
import random

class Scheduler:
  def get_next_time(self, nugget):
    # return the recommended next time this
    # nugget should be tested, based on its
    # past history of test results
    if not nugget.test_results:
      return 0.0

    tr0 = nugget.test_results[0]
    if tr0[1] < 0.5:  # forgotten, goes into short term review
      res = tr0[1] * 3 * 60 + 45
    else:
      # remembered
      int, deg = self._get_last_interval(nugget)
      if int == 0.0:
        # first time remembered
        res = ((deg + 0.5)**1.5) * 12 * 3600
      else:
        # seen it before
        if deg >= 0.5:
          # remembered, so boost
          res = int * (deg + 0.6)**1.7 - 0.1
        else:
          # forgot, so reduce
          res = int * (deg + 0.2)
    # slightly perturb so not all tests come together
    res += random.uniform(-1,1) * 0.1 * res
    return tr0[0] + res 

  def _get_last_interval(self, nugget):
    assert nugget.test_results
    assert nugget.test_results[0][1] >= 0.5
    int = 0.0
    deg = nugget.test_results[0][1]
    if len(nugget.test_results) > 1:
      # find first remembered test since last test, then interval 
      # is time between that and previous 
      # test.  If no time that we've remembered, this 
      # is actually our first time learning it, so interval is 0.0
      for t in range(1, len(nugget.test_results)):
        if nugget.test_results[t][1] >= 0.5:
          int = nugget.test_results[t-1][0] - nugget.test_results[t][0]
          deg = nugget.test_results[t-1][1]
          break
    return int, deg


class KnowledgeOrganizer:
  def __init__(self):
    self._root_collection = None 
    self._known_nuggets = [] # heap of nuggets already tested
    self._new_nuggets = []   # list of nuggets not learned yet
    self._nugget_under_test = None
    
  def load(self, root_name):
    # create KnowledgeCollection rooted at root and load it
    self._root_collection = KnowledgeCollection(root_name)
    # now initialize our data structures
    self._known_nuggets = self.filter_nuggets(lambda n: n.next_test != 0.0)
    heapq.heapify(self._known_nuggets)
    self._new_nuggets = self.filter_nuggets(lambda n: n.next_test == 0.0)
    self._nugget_under_test = None

  def save(self):
    # save out the root KnowledgeCollection
    self._root_collection.save()

  def get_next_nugget(self):
    if self._nugget_under_test:
      pass
    elif self._known_nuggets and self._known_nuggets[0].next_test <= time.time():
      self._nugget_under_test = heapq.heappop(self._known_nuggets)
    elif self._new_nuggets:
      self._nugget_under_test = self._new_nuggets.pop(0)
    if self._nugget_under_test:
      return self._nugget_under_test.filename
    else:
      return ''

  def nugget_test_completed(self, degree_of_remembering):
    assert self._nugget_under_test
    self._nugget_under_test.new_result(degree_of_remembering) 
    self._nugget_under_test.save()
    heapq.heappush(self._known_nuggets, self._nugget_under_test)
    self._nugget_under_test = None

  # return list of results of func of 1 arg applied to all 
  # nuggets in memory
  def map_nuggets(self, func):
    return self._map_nuggets(func, self._root_collection)

  def _map_nuggets(self, func, collection):
    res = []
    res += map(func, collection.nuggets)
    for c in collection.collections:
      res += self._map_nuggets(func, c)
    return res

  def filter_nuggets(self, func):
    return self._filter_nuggets(func, self._root_collection)
 
  def _filter_nuggets(self, func, collection):
    res = []
    res += filter(func, collection.nuggets)
    for c in collection.collections:
      res += self._filter_nuggets(func, c)
    return res

    
    

    

class KnowledgeCollection:
  _nugre = '\.nug$'
  def __init__(self, filename):
    self.filename = filename  # absolute path
    self.collections = []  # list of KnowledgeCollections
    self.nuggets = []  # list of KnowledgeNuggets
    self.load()

  def load(self):
    # look for all .nug directories in my directory, and load them (unpickle)
    # look for all collection directories in my directory, and load them
    root, dirs, files = os.walk(self.filename).next()
    dirs.sort()
    for d in dirs:
      dn = os.path.join(root, d)
      if re.search(self._nugre, d):
        fn = os.path.join(dn, KnowledgeNugget.nugname)
        if os.path.isfile(fn):
          nug = pickle.load(file(fn, 'r'))
        else:
          nug = KnowledgeNugget()
        nug.filename = dn
        self.nuggets.append(nug)
      else:
        self.collections.append(KnowledgeCollection(dn))

  def save(self):
    for n in self.nuggets:
      n.save()
    for c in self.collections:
      c.save()


class KnowledgeNugget:
  nugname = 'data.smx'
  def __init__(self):
    self.filename = ''  # absolute filename path (only dirictory up to *.nug)
    # probability 0.0 to 1.0, typically 0.9
    self.desired_retention = 0.9
    # absolute time() of the next scheduled test for this item, or 0
    # if a new item
    self.next_test = 0.0
    # test_results is a list of tuples,
    # each tuple is a float of the time and a retention_degree
    # from 0.0 to 1.0
    self.test_results = []

  def save(self):
    f = file(os.path.join(self.filename, self.nugname), 'w')
    pickle.dump(self, f)
    f.close() # explicitly close because we want this updated now in case of crash

  def __cmp__(self, other):
    if self.next_test < other.next_test:
      return -1
    elif self.next_test > other.next_test:
      return 1
    else:
      return 0

  def new_result(self, degree_of_remembering):
    self.test_results.insert(0, (time.time(), degree_of_remembering))
    self.next_test = Scheduler().get_next_time(self)
    

class UI:
  def run(self):
    o = KnowledgeOrganizer()
    o.load('/home/rbrewer/python/supermemo/data/bible')
#    print [(n.next_test, n.filename) for n in o._known_nuggets]
#    print [(n.next_test, n.filename) for n in o._new_nuggets]
    while True:
      fn = o.get_next_nugget()
      if fn == '':
        break
      print
      print 'question: ', file(os.path.join(fn, 'q.txt'), 'r').read()
      print
      print 'press enter to see answer'
      raw_input()
      print 'answer: ', file(os.path.join(fn, 'a.txt'), 'r').read()
      print
      while True:
        print 'please rate how well you did: ==> 0 1 2  3 4 5 <=='
        rating = input()
        if 0 <= rating <= 5:
          break
      deg = rating * 20.0 / 100
      o.nugget_test_completed(deg)
    print 'all done'  
    o.save()
      


UI().run()

